@@ -14,34 +14,6 @@ import thingy.git as git
 import thingy.colour as colour
 
 ################################################################################
-# TODO List:
-#
-# DONE: ***MUST use relative paths in config file, or we can't store in git and clone and use somewhere else!***
-# DONE: / Output name of each working tree as it is processed as command sits there seeming to do nothing otherwise.
-# DONE: Better error-handling - e.g. continue/abort option after failure in one working tree
-# DONE: Currently doesn't handle single letter options in concatenated form - e.g. -dv
-# DONE: Don't save the configuration on exit if it hasn't changed
-# DONE: Don't use a fixed list of default branch names
-# DONE: If the config file isn't in the current directory then search up the directory tree for it but run in the current directory
-# DONE: Use the configuration file
-# DONE: When specifying list of working trees, if name doesn't contain '/' prefix it with '*'?
-# DONE: command to categorise working trees then command line filter to only act on working trees in category (in addition to other filtering options) - +tag <TAG> command tags all selected working trees and updates the configuration, +untag <TAG> command to remove tags in the same way
-# DONE: init function
-# DONE Consistent colours in output
-# DONE:+run command to do things other than git commands
-# DONE: Command that takes partial working tree name and either returns full path or pops up window to autocomplete until single match found - returns list of paths you can pipe into fzf
-#
-# NOPE: Clone to have option to update - as discussed, should be part of init
-# NOPE: Dry-run option - just pass the option to the Git command
-# NOPE: Is it going to be a problem if the same repo is checked out twice or more in the same workspace - user problem
-# NOPE: Pull/fetch - only output after running command and only if something updated - nope, git commands should run as-is without interference
-# NOPE: Switch to tomlkit
-# NOPE: Use PathLib - Don't really see the need
-# NOPE: When we have something with multiple matches display a menu for user to select the one that they one - make it a library routine so can be used, for instance, for branch selection - Use fzf
-# NOPE: Use pygit2 directly - don't see the need
-# NOPE: .init option '--update' to update the configuration file with new working trees and remove ones that are no longer there - init does the removal, and added the '+add' command instead
-# NOPE: Ability to read default configuration options from ~/.config/thingy/multigit.rc - these are insert before argv[1] before argparse called - limited use for this
-#
 # TODO: 2. If run in a subdirectory, only process working trees in that tree (or have an option to do so, or an option _not_ to do so; --all)
 # TODO: 2. select_git_repos() and +dir should use consist way of selecting repos if possible
 # TODO: 3 .init option '--set-default' to update the default branch to the current one for specified working trees
@@ -87,16 +59,17 @@ options:
   --path, -C PATH       Run as if the command was started in PATH instead of the current working directory
 
 Sub-commands:
-  {+clone, +init, +dir, +config, +list, +run, +add, GIT_COMMAND}
+  {+clone, +init, +dir, +config, +list, +run, +add, +update, GIT_COMMAND}
     +clone REPO {BRANCH} Clone a repo containing a multigit configuration file, then clone all the child repos and check out the default branch in each
     +init                Build or update the configuration file using the current branch in each repo as the default branch
-    +config              Return the name and location of the configuration file
-    +dir                 Return the location of a working tree, given the repo name, or if no parameter specified, the root directory of the multigit tree
-    +list                Return a list of the top level directories of each of the Git repos
+    +config              Output the name and location of the configuration file
+    +dir                 Output the location of a working tree, given the repo name, or if no parameter specified, the root directory of the multigit tree
+    +list                Output a list of the top level directories of each of the Git repos
     +tag TAG             Apply a configuration tag to repos filtered by the command line options (list configuration tags if no parameter specified)
     +untag TAG           Remove a configuration tag to repos filtered by the command line options
     +run COMMAND         Run the specified command in repos filtered by the command line options
     +add REPO DIR        Clone REPO into the DIR directory and add it to the multigit configuration
+    +update              Clone any repos in the current configuration that do not have working trees
     GIT_COMMAND          Any git command, including options and parameters - this is then run in all specified working trees
 
 """
@@ -168,6 +141,32 @@ def relative_repo_path(args, relative_path=''):
 
 ################################################################################
 
+def safe_clone(args, location, origin):
+    """If location exists then fail with an error if it isn't a directory, or
+    or is a directory, but isn't a working tree, or is a working tree for a
+    different Git repo than remote otherwise do nothing.
+    If it doesn't exist, then clone the specified remote there."""
+
+    if os.path.exists(location):
+        if not os.path.isdir(os.path.join(location, '.git')):
+            colour.error(f'"[BLUE:{location}]" already exists and is not a Git working tree', prefix=True)
+
+        remotes = git.remotes(path=location)
+
+        for remote in remotes:
+            if origin == remotes[remote]:
+                break
+        else:
+            colour.error(f'"[BLUE:{location}]" already exists and was not cloned from [BLUE:{origin}]', prefix=True)
+
+    else:
+        if not args.quiet:
+            colour.write(f'Cloning [BOLD:{origin}] into [BLUE:{location}]')
+
+        git.clone(origin, working_tree=location)
+
+################################################################################
+
 def find_configuration(default_config_file):
     """If the configuration file name has path elements, try and read it, otherwise
        search up the directory tree looking for the configuration file.
@@ -311,36 +310,37 @@ def branch_name(name, default_branch):
 
 ################################################################################
 
-def add_new_repo(args, config, repo):
+def add_new_repo(args, config, repo_path, default_branch=None):
     """Add a new configuration entry containing the default branch, remote origin
-    (if there is one) and name"""
+    (if there is one), name and default branch"""
 
-    abs_repo_path = absolute_repo_path(args, repo)
+    abs_repo_path = absolute_repo_path(args, repo_path)
 
-    added = repo not in config
+    added = repo_path not in config
 
-    config[repo] = {}
-
-    default_branch = git.branch(path=abs_repo_path)
+    config[repo_path] = {}
 
     if not default_branch:
-        colour.error(f'Unable to determine default branch in [BLUE:{repo}]', prefix=True)
+        default_branch = git.branch(path=abs_repo_path)
 
-    config[repo]['default branch'] = default_branch
+    if not default_branch:
+        colour.error(f'Unable to determine default branch in [BLUE:{abs_repo_path}]', prefix=True)
+
+    config[repo_path]['default branch'] = default_branch
 
     remote = git.remotes(path=abs_repo_path)
 
     if 'origin' in remote:
-        config[repo]['origin'] = remote['origin']
-        config[repo]['repo name'] = os.path.basename(remote['origin']).removesuffix('.git')
+        config[repo_path]['origin'] = remote['origin']
+        config[repo_path]['repo name'] = os.path.basename(remote['origin']).removesuffix('.git')
     else:
-        config[repo]['repo name'] = os.path.basename(repo)
+        config[repo_path]['repo name'] = os.path.basename(repo_path)
 
     if not args.quiet:
         if added:
-            colour.write(f'Added [BLUE:{repo}] with default branch [BLUE:{default_branch}]')
+            colour.write(f'Added [BLUE:{repo_path}] with default branch [BLUE:{default_branch}]')
         else:
-            colour.write(f'Reset [BLUE:{repo}] with default branch [BLUE:{default_branch}]')
+            colour.write(f'Reset [BLUE:{repo_path}] with default branch [BLUE:{default_branch}]')
 
 ################################################################################
 
@@ -428,14 +428,14 @@ def mg_init(args, config, console):
     # Search for .git directories and add any that aren't already in the configuration
 
     repo_list = []
-    for repo in find_working_trees(args):
+    for repo_dir in find_working_trees(args):
         if not args.quiet:
-            show_progress(console.columns, repo)
+            show_progress(console.columns, repo_dir)
 
-        repo_list.append(repo)
+        repo_list.append(repo_dir)
 
-        if repo not in config:
-            add_new_repo(args, config, repo)
+        if repo_dir not in config:
+            add_new_repo(args, config, repo_dir)
 
     if not args.quiet:
         colour.write(cleareol=True)
@@ -476,22 +476,9 @@ def mg_add(args, config, console):
     repo = args.parameters[0]
     location = args.parameters[1]
 
-    if os.path.exists(location):
-        if os.path.isdir(os.path.join(location, '.git')):
-            remotes = git.remotes(path=location)
+    # Attempt to clone it
 
-            for remote in remotes:
-                if repo.endswith('.git') and repo == remotes[remote]:
-                    break
-                elif f'{repo}.git' == remotes[remote]:
-                    break
-            else:
-                colour.error(f'"[BLUE:{location}]" already exists and was not cloned from [BLUE:{repo}]', prefix=True)
-        else:
-            colour.error(f'"[BLUE:{location}]" already exists and is not a Git working tree', prefix=True)
-
-    else:
-        git.clone(repo, working_tree=location)
+    safe_clone(args, repo, location)
 
     # Add to the configuration
 
@@ -503,6 +490,26 @@ def mg_add(args, config, console):
 
 ################################################################################
 
+def mg_update(args, config, console):
+    """Clone any repos in the current configuration that do not have working trees
+       Similar to the '+init' command except that it updates an existing multigit
+       tree rather than creating one from scratch."""
+
+    _ = console
+
+    # Don't allow pointless options
+
+    if args.modified or args.branched or args.tag:
+        colour.error('The "[BOLD:--tag]", "[BOLD:--modified]" and "[BOLD:--branched]" options cannot be used with the "[BOLD:+update]" subcommand', prefix=True)
+
+    # Now iterate through the repos, cloning any that don't already have working trees
+
+    for repo in select_git_repos(args, config):
+        if repo.name != '.':
+            safe_clone(args, repo.name, repo['origin'])
+
+################################################################################
+
 def mg_dir(args, config, console):
     """Return the location of a working tree, given the name, or the root directory
        of the tree if not
@@ -813,6 +820,7 @@ COMMANDS = {
     'list': mg_list,
     'run': mg_run,
     'add': mg_add,
+    'update': mg_update,
 }
 
 def main():
